from math import factorial

import numpy as np
import matplotlib.pyplot as plt

from pysprint.utils import pad_with_trailing_zeros


__all__ = [
    "transform_cf_params_to_dispersion",
    "transform_lmfit_params_to_dispersion",
    "plot_phase",
]


def transform_cf_params_to_dispersion(popt, drop_first=True, dof=0):
    if drop_first:
        popt = popt[1:]

    _disp = popt

    for idx in range(len(popt)):
        _disp[idx] = popt[idx] * factorial(idx + dof)

    _disp = pad_with_trailing_zeros(_disp, 5)
    return _disp, np.zeros(5)


def transform_lmfit_params_to_dispersion(popt, popt_std, drop_first=True, dof=1):
    if drop_first:
        popt, popt_std = popt[1:], popt_std[1:]

    _disp, _disp_std = popt, popt_std

    for idx in range(len(popt)):
        _disp[idx] = popt[idx] * factorial(idx + dof)
        if popt_std[idx] is not None:
            _disp_std[idx] = popt_std[idx] * factorial(idx + dof)
        else:
            _disp_std[idx] = 0
    _disp = pad_with_trailing_zeros(popt, 5)
    _disp_std = pad_with_trailing_zeros(popt_std, 5)
    return _disp, _disp_std


# DEPRECATE THIS
def plot_phase(
    x,
    y,
    bf,
    *,
    bf_fallback=None,
    figsize=(7, 7),
    window_title=None,
    show_labels=True,
    dataset_color="k",
    fit_color="r",
    grid=True
):

    if window_title is None:
        window_title = "Phase"

    fig = plt.figure(figsize=figsize)
    fig.canvas.set_window_title(window_title)
    plt.plot(x, y, c=dataset_color, label="Dataset")
    try:
        plt.plot(x, bf, c=fit_color, linestyle="dashed", label="Best fit")
    except ValueError:
        if bf_fallback is None:
            pass
        else:
            plt.plot(
                x, bf_fallback, c=fit_color, linestyle="dashed", label="Best fit",
            )
    if show_labels:
        plt.legend()
    if grid:
        plt.grid()
    plt.show()



from unittest.mock import patch

import pytest
import numpy as np

from pysprint import Generator, WFTMethod


@pytest.mark.slow
@patch('matplotlib.pyplot.show')
def test_basic(mck):
    g = Generator(
        1,
        3,
        2,
        3000,
        GDD=400,
        TOD=4000,
        FOD=4000,
        QOD=50000,
        pulse_width=5,
        resolution=0.01
    )

    g.generate()

    f = WFTMethod(*g.data)
    f.add_window_linspace(1.25, 2.75, 350, fwhm=0.017)

    d, _, _ = f.calculate(reference_point=2, order=5, fastmath=False)

    np.testing.assert_array_almost_equal(d, [3000.19708, 399.94187, 3998.03310, 3991.45904, 49894.99719], decimal=4)

    f.heatmap()
    f.show()
    mck.assert_called()



import os
import unittest
from unittest.mock import patch

import pytest
import numpy as np
import matplotlib.pyplot as plt

from pysprint import MinMaxMethod


class TestEvaluate(unittest.TestCase):
    def setUp(self):
        self.x = np.arange(1, 1000, 1)
        self.y = np.sin(self.x)

    def tearDown(self):
        pass

    @pytest.mark.skipif("TF_BUILD" in os.environ, reason="Azure Pipelines fails this.")
    @patch("matplotlib.pyplot.show")
    def test_edit_session(self, mock_show):
        ifg = MinMaxMethod(self.x, self.y)
        ifg.init_edit_session(engine="cwt")
        ifg.init_edit_session(engine="slope")
        ifg.init_edit_session(engine="normal")
        plt.close("all")
        mock_show.assert_called()

    def test_edit_session2(self):
        ifg = MinMaxMethod(self.x, self.y)
        with self.assertRaises(ValueError):
            ifg.init_edit_session(engine="dfssdfasdf")

    def test_edit_session3(self):
        ifg = MinMaxMethod(self.x, self.y)
        with self.assertRaises(TypeError):
            ifg.init_edit_session(engine="normal", invalidkwarg=3)

    @pytest.mark.skipif("TF_BUILD" in os.environ, reason="Fails on azure.")
    @patch("matplotlib.pyplot.show")
    def test_edit_session4(self, mock_show):
        ifg = MinMaxMethod(self.x, self.y)
        ifg.init_edit_session()
        plt.close("all")
        mock_show.assert_called()


if __name__ == "__main__":
    unittest.main()



import warnings

import matplotlib # noqa
import matplotlib.pyplot as plt

from pysprint.utils.misc import run_from_ipython


warnings.filterwarnings("ignore", message="invalid value encountered in sqrt")
warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")


def interactive(figsize=(15, 5), backend="Qt5Agg"):
    plt.ion()
    plt.rcParams["figure.figsize"] = figsize
    try:
        from IPython import get_ipython

        ipython = get_ipython()
    except ImportError:
        import IPython.ipapi

        ipython = IPython.ipapi.get()
    try:
        ipython.magic("matplotlib qt")
    except AttributeError:
        pass
    else:
        pass

    try:
        plt.switch_backend(backend)
    except AttributeError:
        warnings.warn(
            "You should manually set a suitable matplotlib backend, "
            "e.g. matplotlib.use('Qt5Agg') to enable interactive plots."
        )


__version__ = "0.12.1"
__author__ = "Leéh Péter"

from .api import *
from .utils import print_info



import os
import re
from copy import copy
from functools import wraps, lru_cache

import numpy as np
from scipy.interpolate import interp1d
import scipy.stats as st

__all__ = [
    "unpack_lmfit",
    "find_nearest",
    "_handle_input",
    "print_disp",
    "fourier_interpolate",
    "between",
    "get_closest",
    "run_from_ipython",
    "calc_envelope",
    "measurement",
    "_maybe_increase_before_cwt",
    "pad_with_trailing_zeros",
    "mutually_exclusive_args",
    "lazy_property",
    "inplacify"
]

_inplace_doc = """\n\tinplace : bool, optional
            Whether to apply the operation on the dataset in an "inplace" manner.
            This means if inplace is True it will apply the changes directly on
            the current dataset and returns None. If inplace is False, it will
            leave the current object untouched, but returns a copy of it, and
            the operation will be performed on the copy. It's useful when
            chaining operations on a dataset.\n\n\t"""


def _has_parameter_section(method):
    try:
        return "Parameters" in method.__doc__
    except TypeError:
        return False


def update_doc(method, doc):
    if _has_parameter_section(method):
        newdoc = _build_doc(method, doc)
        method.__doc__ = newdoc
    else:
        newdoc = """\n\tParameters
        ----------\n\tinplace : bool, optional
            Whether to apply the operation on the dataset in an "inplace" manner.
            This means if inplace is True it will apply the changes directly on
            the current dataset and returns None. If inplace is False, it will
            leave the current object untouched, but returns a copy of it, and
            the operation will be performed on the copy. It's useful when
            chaining operations on a dataset.\n\n\t"""

        nodoc_head = (f"Docstring automatically created for {method.__name__}. "
                      "Parameter list may not be complete.\n")
        if method.__doc__ is not None:
            method.__doc__ += newdoc
        else:
            method.__doc__ = nodoc_head + newdoc
        return


def _build_doc(method, param):
    patt = r"(\w+(?=\s*[-]{4,}[^/]))"  # finding sections
    splitted_doc = re.split(patt, method.__doc__)
    try:
        target = splitted_doc.index("Parameters") + 1
    except ValueError:
        return method.__doc__

    splitted_doc[target] = splitted_doc[target].rstrip() + param

    return ''.join(_ for _ in splitted_doc if _ is not None)


def inplacify(method):
    update_doc(method, _inplace_doc)

    @wraps(method)
    def wrapper(self, *args, **kwds):
        inplace = kwds.pop("inplace", True)
        if inplace:
            method(self, *args, **kwds)
        else:
            return method(copy(self), *args, **kwds)

    return wrapper


# https://stackoverflow.com/a/54487188/11751294
def mutually_exclusive_args(keyword, *keywords):
    """"
    Decorator to restrict the user to specify exactly one of the given parameters.
    Often used for std and fwhm for Gaussian windows.
    """
    keywords = (keyword,) + keywords

    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            if sum(k in keywords for k in kwargs) != 1:
                raise TypeError(
                    "You must specify exactly one of {}.".format(" and ".join(keywords))
                )
            return func(*args, **kwargs)

        return inner

    return wrapper


def lazy_property(f):
    return property(lru_cache()(f))


def _maybe_increase_before_cwt(y, tolerance=0.05):
    y = np.asarray(y)
    value = y.min()
    if value - tolerance >= 0:
        return False
    if value < 0 < np.abs(value) - tolerance:
        return True
    return True


def calc_envelope(x, ind, mode="u"):
    """Source: https://stackoverflow.com/a/39662343/11751294
    """
    x_abs = np.abs(x)
    if mode == "u":
        loc = np.where(np.diff(np.sign(np.diff(x_abs))) < 0)[0] + 1
    elif mode == "l":
        loc = np.where(np.diff(np.sign(np.diff(x_abs))) > 0)[0] + 1
    else:
        raise ValueError("mode must be u or l.")
    peak = x_abs[loc]
    envelope = np.interp(ind, loc, peak)
    return envelope, peak, loc


def run_from_ipython():
    """
    Detect explicitly if code is run inside Jupyter.
    """
    try:
        __IPYTHON__
        if any("SPYDER" in name for name in os.environ):
            return False
        return True
    except NameError:
        return False


def get_closest(x_val, y_val, x_array, y_array):
    idx = np.argmin((np.hypot(x_array - x_val, y_array - y_val)))
    value = x_array[idx]
    return value, y_array[idx], idx


def between(val, except_around):
    if except_around is None:
        return False
    elif len(except_around) != 2:
        raise ValueError(
            f"Invalid interval. Try [start, end] instead of {except_around}"
        )
    else:
        lower = float(min(except_around))
        upper = float(max(except_around))
    if upper >= val >= lower:
        return True
    return False


def unpack_lmfit(r):
    dispersion, dispersion_std = [], []
    for name, par in r:
        dispersion.append(par.value)
        dispersion_std.append(par.stderr)
    return dispersion, dispersion_std


def measurement(array, confidence=0.95, silent=False):
    """
    Give the measurement results with condifence interval
    assuming the standard deviation is unknown.

    Parameters
    ----------

    array : ndarray
        The array containing the measured values

    confidence : float, optional
        The desired confidence level. Must be between 0 and 1.

    silent : bool, optional
        Whether to print results immediately. Default is `False`.

    Returns
    -------

    mean: float
        The mean of the given array

    conf: tuple-like (interval)
        The confidence interval

    Examples
    ---------
    >>> import numpy as np
    >>> from pysprint.utils import measurement
    >>> a = np.array([123.783, 121.846, 122.248, 125.139, 122.569])
    >>> mean, interval = measurement(a, 0.99)
    123.117000 ± 2.763022
    >>> mean
    123.117
    >>> interval
    (120.35397798230359, 125.88002201769642)

    Notes
    -----

    I decided to print the results immediately, because people often don't use
    it for further code. Of course, they are also returned if needed.
    """
    mean = np.mean(array)
    conf = st.t.interval(confidence, len(array) - 1, loc=mean, scale=st.sem(array))
    if not silent:
        print(f"{mean:5f} ± {(mean - conf[0]):5f}")
    return mean, conf


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(value - array)).argmin()
    return array[idx], idx


def _handle_input(x, y, ref, sam):
    """
    Instead of handling the inputs in every function, there is this private method.

    Parameters
    ----------

    x: array-like
    x-axis data

    y: array-like
    y-axis data

    ref, sam: array-like
    reference and sample arm spectrum evaluated at x

    Returns
    -------
    x: array-like
    unchanged x data

    Ydata: array-like
    the transformed y data

    """
    if len(x) == 0:
        raise ValueError("No values for x.")

    if len(y) == 0:
        raise ValueError("No values for y.")

    if (len(x) > 0) and (len(ref) > 0) and (len(sam) > 0):
        y_data = (y - ref - sam) / (2 * np.sqrt(ref * sam))
    elif (len(ref) == 0) or (len(sam) == 0):
        y_data = y
    else:
        raise TypeError("Input types are wrong.\n")
    return np.asarray(x), np.asarray(y_data)


def print_disp(f):
    @wraps(f)
    def wrapping(*args, **kwargs):
        disp, disp_std, stri = f(*args, **kwargs)
        labels = ("GD", "GDD", "TOD", "FOD", "QOD", "SOD")
        disp = np.trim_zeros(disp, "b")
        disp_std = disp_std[: len(disp)]
        for i, (label, disp_item, disp_std_item) in enumerate(
                zip(labels, disp, disp_std)
        ):
            if run_from_ipython():
                from IPython.display import display, Math

                display(
                    Math(f"{label} = {disp_item:.5f} ± {disp_std_item:.5f} fs^{i + 1}")
                )
            else:
                print(f"{label} = {disp_item:.5f} ± {disp_std_item:.5f} fs^{i + 1}")
        return disp, disp_std, stri

    return wrapping


def fourier_interpolate(x, y):
    """ Simple linear interpolation for FFTs"""
    xs = np.linspace(x[0], x[-1], len(x))
    intp = interp1d(x, y, kind="linear", fill_value="extrapolate")
    ys = intp(xs)
    return xs, ys


def pad_with_trailing_zeros(array, shape):
    """
    Pad an array with trailing zeros to be the desired shape
    """
    c = shape - len(array)
    return np.pad(array, pad_width=(0, c), mode="constant")



import sys
import warnings

import numpy as np
import matplotlib.pyplot as plt

from pysprint.core.phase import Phase
from pysprint.utils import PySprintWarning
from pysprint.core.fft_tools import find_center, find_roi
from pysprint.core.methods._fft import FFTMethod
from pysprint.core.evaluate import gaussian_window
from pysprint.utils.misc import mutually_exclusive_args, lazy_property, find_nearest


class Window:
    """
    Basic class that implements functionality related to Gaussian
    windows with caching the y values.
    """

    def __init__(self, x, center, fwhm, order=2):
        self.x = x
        self.center = center
        self.fwhm = fwhm
        self.order = order

    @lazy_property
    def y(self):
        return gaussian_window(self.x, self.center, self.fwhm, self.order)

    @classmethod
    def from_std(cls, x, center, std, order=2):
        _fwhm = std * 2 * np.log(2) ** (1 / order)
        return cls(x, center, _fwhm, order)

    def __repr__(self):
        return f"Window(center={self.center:.5f}, fwhm={self.fwhm}, order={self.order})"

    def plot(self, ax=None, scalefactor=1, zorder=90, **kwargs):
        """
        Plot the window.

        Parameters
        ----------

        ax : axis
            The axis to plot on. If not given, plot on the last axis.

        scalefactor : float, optional
            Number describing how much a given should be scaled up ONLY
            for visibility.

        zorder : float, optional
            The drawing order of artists is determined by their zorder attribute, which is
            a floating point number. Artists with higher zorder are drawn on top. You can
            change the order for individual artists by setting their zorder. The default
            value depends on the type of the Artist.
        """
        if ax is None:
            ax = plt
        ax.plot(self.x, self.y * scalefactor, zorder=zorder, **kwargs)


class WFTMethod(FFTMethod):
    """Basic interface for Windowed Fourier Transform Method."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.window_seq = {}
        self.found_centers = {}
        self.secondary_centers = {}
        self.tertiary_centers = {}
        self.GD = None
        self.cachedlen = 0
        self.X_cont = np.array([])
        self.Y_cont = np.array([])
        self.Z_cont = np.array([])
        self.fastmath = True

    @mutually_exclusive_args("std", "fwhm")
    def add_window(self, center, std=None, fwhm=None, order=2):
        if not np.min(self.x) <= center <= np.max(self.x):
            raise ValueError(
                f"Cannot add window at {center}, because "
                f"it is out of the dataset's range (from {np.min(self.x):.3f} to {np.max(self.x):.3f})."
            )
        if std:
            window = Window.from_std(
                self.x, center=center, std=std, order=order
            )
        else:
            window = Window(self.x, center=center, fwhm=fwhm, order=order)
        self.window_seq[center] = window

    @property
    def windows(self):
        return self.window_seq

    @property
    def centers(self):
        return self.window_seq.keys()

    @mutually_exclusive_args("std", "fwhm")
    def add_window_arange(
        self, start, stop, step, std=None, fwhm=None, order=2
    ):
        """
        Build a window sequence of given parameters to apply on ifg.
        Works similar to numpy.arange.
        """
        arr = np.arange(start, stop, step)
        for cent in arr:
            if std:
                self.add_window(center=cent, std=std, order=order)
            else:
                self.add_window(center=cent, fwhm=fwhm, order=order)

    @mutually_exclusive_args("std", "fwhm")
    def add_window_linspace(
        self, start, stop, num, std=None, fwhm=None, order=2
    ):
        """
        Build a window sequence of given parameters to apply on ifg.
        Works similar to numpy.linspace.
        """
        arr = np.linspace(start, stop, num)
        for cent in arr:
            if std:
                self.add_window(center=cent, std=std, order=order)
            else:
                self.add_window(center=cent, fwhm=fwhm, order=order)

    @mutually_exclusive_args("std", "fwhm")
    def add_window_geomspace(
        self, start, stop, num, std=None, fwhm=None, order=2
    ):
        """
        Build a window sequence of given parameters to apply on ifg.
        Works similar to numpy.geomspace.
        """
        arr = np.geomspace(start, stop, num)
        for cent in arr:
            if std:
                self.add_window(center=cent, std=std, order=order)
            else:
                self.add_window(center=cent, fwhm=fwhm, order=order)

    #  TODO : scale them up for visibility
    def view_windows(self, ax=None, maxsize=80, **kwargs):
        """
        Gives a rough view of the different windows along with the ifg.
        """
        winlen = len(self.window_seq)
        ratio = winlen % maxsize
        if winlen > maxsize:
            warnings.warn(
                "Image seems crowded, displaying only a sequence of the given windows.",
                PySprintWarning
            )
            for i, (_, val) in enumerate(self.window_seq.items()):
                if i % ratio == 0:
                    val.plot(ax=ax, scalefactor=np.max(self.y) * .75, **kwargs)
        else:
            for _, val in self.window_seq.items():
                val.plot(ax=ax, scalefactor=np.max(self.y) * .75, **kwargs)
        self.plot(ax=ax)

    def remove_all_windows(self):
        self.window_seq.clear()

    def reset_state(self):
        self.remove_all_windows()
        self.found_centers.clear()
        self.X_cont = np.array([])
        self.Y_cont = np.array([])
        self.Z_cont = np.array([])
        self.GD = None
        self.cachedlen = 0
        self.fastmath = True

    def remove_window_at(self, center):
        if center not in self.window_seq.keys():
            c = find_nearest(
                np.fromiter(self.window_seq.keys(), dtype=float), center
            )
            raise ValueError(
                f"There is no window with center {center}. "
                f"Did you mean {c[0]}?"
            )
        self.window_seq.pop(center, None)

    def remove_window_interval(self, start, stop):
        pass

    # TODO : Add parameter to describe how many peaks we are looking for..
    def calculate(
            self,
            reference_point,
            order,
            show_graph=False,
            silent=False,
            force_recalculate=False,
            fastmath=True,
            usenifft=False
    ):
        if len(self.window_seq) == 0:
            raise ValueError("Before calculating a window sequence must be set.")

        if self.cachedlen != len(self.window_seq) or fastmath != self.fastmath:
            force_recalculate = True
        self.fastmath = fastmath
        if force_recalculate:
            self.found_centers.clear()
            self.build_GD(silent=silent, fastmath=fastmath, usenifft=usenifft)
        if self.GD is None:
            self.build_GD(silent=silent, fastmath=fastmath, usenifft=usenifft)

        self.cachedlen = len(self.window_seq)

        if order == 1:
            raise ValueError("Cannot fit constant function to data. Order must be in [2, 5].")

        d, ds, fr = self.GD._fit(
            reference_point=reference_point, order=order
        )
        if show_graph:
            self.GD.plot()
        return d, ds, fr

    def build_GD(self, silent=False, fastmath=True, usenifft=False):
        self.fastmath = fastmath
        self._apply_window_sequence(silent=silent, fastmath=fastmath, usenifft=usenifft)
        self._clean_centers()
        delay = np.fromiter(self.found_centers.keys(), dtype=float)
        omega = np.fromiter(self.found_centers.values(), dtype=float)
        self.GD = Phase(delay, omega, GD_mode=True)
        return self.GD

    def _predict_ideal_window_fwhm(self):
        pass

    def _apply_window_sequence(
            self, silent=False, fastmath=True, usenifft=False, errors="ignore"
    ):
        winlen = len(self.window_seq)

        if not fastmath:
            # here we setup the shape for the Z array because
            # it is much faster than using np.append in every iteration
            _x, _y, _, _ = self._safe_cast()
            _obj = FFTMethod(_x, _y)
            _obj.ifft(usenifft=usenifft)
            x, y = find_roi(_obj.x, _obj.y)
            yshape = y.size
            xshape = len(self.window_seq)
            self.Z_cont = np.empty(shape=(yshape, xshape))

        for idx, (_center, _window) in enumerate(self.window_seq.items()):
            _x, _y, _, _ = self._safe_cast()
            _obj = FFTMethod(_x, _y)
            _obj.y *= _window.y
            _obj.ifft(usenifft=usenifft)
            x, y = find_roi(_obj.x, _obj.y)
            if not fastmath:
                if self.Y_cont.size == 0:  # prevent allocating it in every iteration
                    self.Y_cont = np.array(x)
                self.Z_cont[:, idx] = y
            try:
                centx, _ = find_center(x, y)
                self.found_centers[_center] = centx
            except ValueError as err:
                if errors == "ignore":
                    self.found_centers[_center] = None
                else:
                    raise err
            if not silent:
                sys.stdout.write('\r')
                j = (idx + 1) / winlen
                sys.stdout.write("Progress : [%-30s] %d%%" % ('=' * int(30 * j), 100 * j))
                sys.stdout.flush()

    def _clean_centers(self, silent=False):
        dct = {k: v for k, v in self.found_centers.items() if v is not None}
        self.found_centers = dct
        winlen = len(self.window_seq)
        usefullen = len(self.found_centers)

        if not silent:
            if winlen != usefullen:
                print(
                    f"\nIn total {winlen-usefullen} out of {winlen} datapoints "
                    f"were thrown away due to ambiguous peak positions."
                )

    def errorplot(self, *args, **kwargs):
        try:
            getattr(self.GD, "errorplot", None)(*args, **kwargs)
        except TypeError:
            raise ValueError("Must calculate before plotting errors.")

    @property
    def get_GD(self):
        if self.GD is not None:
            return self.GD
        raise ValueError("Must calculate GD first.")

    @property
    def errors(self):
        return getattr(self.GD, "errors", None)

    # TODO : Integrate this into _apply_window_sequence and add matplotlib dispatcher
    def _prepare_element(self, center):
        if center not in self.window_seq.keys():
            raise ValueError(
                f"Window with center {center} cannot be found."
            )
        _x, _y, _, _ = self._safe_cast()
        _obj = FFTMethod(_x, _y)
        _obj.y *= self.window_seq[center].y
        _obj.ifft()
        x, y = find_roi(_obj.x, _obj.y)
        try:
            xx, yy = find_center(x, y)
            _obj.plt.plot(xx, yy, markersize=10, marker="*")
        except ValueError:
            pass
        _obj.plot()
        _obj.show()

    def _construct_heatmap_data(self):
        self.X_cont = np.fromiter(self.window_seq.keys(), dtype=float)

    def heatmap(self, ax=None, levels=None, cmap="viridis", include_ridge=True):
        if self.GD is None:
            raise ValueError("Must calculate first.")

        if self.fastmath:
            raise ValueError(
                "You need to recalculate with `fastmath=False` to plot the heatmap."
            )
        # Only construct if we need to..
        if not (self.Y_cont.size, self.X_cont.size) == self.Z_cont.shape:
            self._construct_heatmap_data()
        if levels is None:
            levels = np.linspace(0, 0.02, 30)
        else:
            if not isinstance(levels, np.ndarray):
                raise ValueError("Expected np.ndarray as levels.")
        if ax is None:
            cs = plt.contourf(
                self.X_cont, self.Y_cont, self.Z_cont, levels=levels, cmap=cmap, extend="both"
            )
        else:
            cs = ax.contourf(
                self.X_cont, self.Y_cont, self.Z_cont, levels=levels, cmap=cmap, extend="both"
            )
        if include_ridge:
            if ax is None:
                plt.plot(*self.GD.data, color='red', label='detected ridge')
            else:
                ax.plot(*self.GD.data, color='red', label='detected ridge')
            plt.legend()
        if ax is None:
            plt.xlabel('Window center [PHz]')
            plt.ylabel('Delay [fs]')
            plt.ylim(None, 1.5 * np.max(self.GD.data[1]))
        else:
            ax.set_autoscalex_on(False)
            ax.set(
                xlabel="Window center [PHz]",
                ylabel="Delay [fs]",
                ylim=(None, 1.5 * np.max(self.GD.data[1]))
            )

    def get_heatmap_data(self):
        """
        Return the data which was used to create the heatmap.

        Returns
        -------

        X_cont : np.ndarray
            The window centers with shape (n,).

        Y_cont : np.ndarray
            The time axis calculated from the IFFT of the dataset with shape (m,).

        Z_cont : np.ndarray
            2D array with shape (m, n) containing the depth information.
        """
        if all([self.Y_cont.size != 0, self.Z_cont.size != 0]):
            self._construct_heatmap_data()
        else:
            raise ValueError("Must calculate with `fastmath=False` before trying to access the heatmap data.")

        return self.X_cont, self.Y_cont, self.Z_cont
